from typing import Optional, List
import dataclasses
from dataclasses import dataclass, field, asdict as dataclassasdict
import uuid

from .types import (
    ArgsDict,
    EnvDict,
)


@dataclass
class Run:
    program: List[str]
    args: Optional[ArgsDict] = field(default_factory=lambda: {})
    env: Optional[EnvDict] = field(default_factory=lambda: {})

    @property
    def argv(self):
        return (
            [f"{k}={v}" for k, v in self.env.items()]
            + self.program
            + [f"--{k}={v}" for k, v in self.args.items()]
        )

    def todict(self):
        return {k: v for k, v in dataclassasdict(self).items()}

    def __str__(self):
        return " ".join(self.argv)

    def __hash__(self):
        return self.id

    @property
    def id(self):
        dict_str = (
            str({k: getattr(self, k) for k in self.fields})
            .encode("utf-8")
            .decode("utf-8")
        )

        return str(uuid.uuid5(uuid.NAMESPACE_DNS, dict_str))[:8]

    @property
    def fields(self):
        return [f.name for f in dataclasses.fields(Run) if not f.name.startswith("_")]



from typing import List
from dataclasses import dataclass, field

from .base import BaseExecutor


@dataclass
class PythonExecutor(BaseExecutor):
    executable: List[str] = field(default_factory=lambda: ["python"])



from typing import List, Optional, Union
from enum import Enum
from dataclasses import dataclass, field

from .base import BaseExecutor


@dataclass
class SlurmExecutor(BaseExecutor):
    executable: List[str] = field(default_factory=lambda: ["sdocker", "python"])
    account: Optional[str] = field(default=None)
    timelimit: Optional[int] = field(default=None)
    gpus: Optional[Union[int, str]] = field(default=None)

    class Env(str, Enum):
        ACCOUNT = "SBATCH_ACCOUNT"
        TIMELIMIT = "HH"
        GPUS = "GPUS"

    def __post_init__(self):
        if self.account is not None:
            self.env[self.Env.ACCOUNT.value] = self.account

        if self.timelimit is not None:
            self.env[self.Env.TIMELIMIT.value] = self.timelimit

        if self.gpus is not None:
            self.env[self.Env.GPUS.value] = self.gpus



from .sweep import Sweep
from .run import Run
from .executors import BaseExecutor
from .providers import BaseProvider, StatelessProvider

__all__ = [
    "__version__",
    "Sweep",
    "Run",
    "BaseExecutor",
    "BaseProvider",
    "StatelessProvider",
]

__version__ = "0.1.2"



from typing import Iterator, Iterable
from enum import Enum
from dataclasses import dataclass

from ..run import Run
from ..types import ArgsDict


@dataclass
class BaseProvider:
    def __call__(self, run: Run, *, sweep=None, **_) -> Run:
        raise NotImplementedError

    def runs(self, config_keys: Iterable[str]) -> Iterator[ArgsDict]:
        raise NotImplementedError


@dataclass
class StatelessProvider(BaseProvider):
    class Env(str, Enum):
        ID = "RUN_ID"
        TAGS = "RUN_TAGS"

    def __call__(self, run: Run, *, sweep=None, **_) -> Run:
        if sweep:
            run.env[self.Env.TAGS.value] = ",".join(sweep.tags)

        run.env[self.Env.ID.value] = run.id

        return run

    def runs(self, _):
        return []



from typing import Iterator, Iterable
import itertools
from functools import reduce
from operator import ior, iand
import polars as pl

from .types import Arg, ArgsMatrix, ArgsDict


def iter_dict(inputs: ArgsMatrix) -> Iterator[ArgsDict]:
    consts = {k: v for k, v in inputs.items() if not isinstance(v, list)}
    inputs = {k: v for k, v in inputs.items() if isinstance(v, list)}

    for values in itertools.product(*inputs.values()):
        yield {**consts, **dict(zip(inputs.keys(), values))}


def ensure_df_columns(
    df: pl.DataFrame, exist_cols: Iterable[str], update: bool = False
) -> pl.DataFrame:
    exist_cols = set(exist_cols)
    cols = set(df.columns)

    missing_cols = exist_cols - cols
    if missing_cols:
        if update:
            df = df.with_columns(**{k: pl.lit(None) for k in missing_cols})
        else:
            msg = ", ".join([f'"{c}"' for c in missing_cols])
            raise ValueError(f"{len(missing_cols)} column(s) missing: {msg}.")

    return df


def ensure_df_compat(*dfs: pl.DataFrame, update: bool = False):
    schemas = tuple(dict(df.schema) for df in dfs)
    cols = set([k for s in schemas for k in s.keys()])

    merged_schema = {}
    for c in cols:
        dtypes = set(
            filter(
                lambda dtype: dtype != pl.datatypes.classes.Null,
                set(s.get(c, pl.datatypes.classes.Null) for s in schemas),
            )
        )

        if len(dtypes) > 1:
            raise ValueError(f'Incompatible dtypes for column "{c}": {dtypes}')

        merged_schema[c] = list(dtypes)[0]

    if update:
        return (
            ensure_df_columns(df, merged_schema.keys(), update=update).cast(
                merged_schema
            )
            for df in dfs
        )

    return dfs


def prepare_df_match_expr(df: pl.DataFrame, match_dict: ArgsMatrix):
    def _expr(k: str, v: Arg):
        if isinstance(v, str):
            ## NOTE: Regex syntax at https://docs.rs/regex/latest/regex/#syntax
            return pl.col(k).str.contains(v, literal=False)
        return pl.col(k) == v

    return reduce(
        ior,
        [
            reduce(iand, [_expr(k, v) for k, v in md.items()])
            for md in iter_dict(match_dict)
        ],
    )



from typing import List, Dict
from dataclasses import dataclass, field


@dataclass
class BaseExecutor:
    executable: List[str] = None
    file: str = None
    env: Dict[str, str] = field(default_factory=lambda: {})

    def __post_init__(self): ...

    @property
    def exec(self) -> List[str]:
        return self.executable + [self.file]



from typing import Dict, Iterable, Union, Tuple, TypeAlias


Arg: TypeAlias = Union[str, int, float]

ArgsDict: TypeAlias = Dict[str, Arg]

EnvDict: TypeAlias = Dict[str, str]

ArgsMatrix: TypeAlias = Dict[str, Union[Arg, Iterable[Arg]]]

IncludeTuple: TypeAlias = Tuple[ArgsMatrix, ArgsDict]

Includes: TypeAlias = Union[IncludeTuple, Iterable[IncludeTuple]]

Excludes: TypeAlias = Union[ArgsMatrix, Iterable[ArgsMatrix]]



from typing import Iterable, Iterator
from tqdm.auto import tqdm
import os
from dataclasses import dataclass, field
from enum import Enum

try:
    import wandb
except ImportError:
    raise ImportError("Missing wandb. Install using `pip install wandb`.")

from .base import BaseProvider
from ..types import ArgsDict
from ..run import Run


@dataclass
class WandBProvider(BaseProvider):
    api_key: str = field(default=os.getenv(wandb.env.API_KEY))
    entity: str = field(default=os.getenv(wandb.env.ENTITY))
    project: str = field(default=os.getenv(wandb.env.PROJECT))
    sweep_id: str = field(default=os.getenv(wandb.env.SWEEP_ID))

    class State(str, Enum):
        FINISHED = "finished"
        FAILED = "failed"
        CRASHED = "crashed"
        RUNNING = "running"

    def __post_init__(self):
        self.api = wandb.Api(
            api_key=self.api_key,
            overrides={
                "entity": self.entity,
                "project": self.project,
            },
        )

    def __call__(self, run: Run, *, sweep=None, **_) -> Run:
        if self.entity:
            run.env[wandb.env.ENTITY] = self.entity

        if self.project:
            run.env[wandb.env.PROJECT] = self.project

        if self.sweep_id:
            run.env[wandb.env.SWEEP_ID] = self.sweep_id

        if sweep:
            from ..sweep import Sweep

            if not isinstance(sweep, Sweep):
                raise ValueError(f"Expected type Sweep. Found {type(sweep).__name__}.")

            run.env[wandb.env.TAGS] = ",".join(sweep.tags)

        run.env[wandb.env.RESUME] = "allow"

        run.env[wandb.env.RUN_ID] = run.id

        return run

    def runs(self, config_keys: Iterable[str]) -> Iterator[ArgsDict]:
        yield from self._get_run_configs(filter_keys=config_keys)

    def _get_runs(
        self, filters: dict = None, per_page: int = 1000, **kwargs
    ) -> Iterator[wandb.apis.public.Run]:
        filters = filters or {}
        if self.sweep_id:
            filters["sweep"] = self.sweep_id

        return self.api.runs(filters=filters, per_page=per_page, **kwargs)

    def _get_unfinished_runs(
        self, filters: dict = None, **kwargs
    ) -> Iterator[wandb.apis.public.Run]:
        filters = {
            **(filters or {}),
            "state": {
                "$nin": [
                    self.State.FINISHED,
                    self.State.RUNNING,
                ]
            },
        }

        return self._get_runs(filters=filters, **kwargs)

    def _get_run_configs(
        self, filter_keys: Iterable[str] = None, **kwargs
    ) -> Iterator[ArgsDict]:
        filter_keys = set(filter_keys or [])
        for run in tqdm(self._get_runs(**kwargs), leave=False):
            cfg = {k: v for k, v in run.config.items()}
            if filter_keys:
                cfg = {k: v for k, v in cfg.items() if k in filter_keys}
            yield cfg



from typing import Iterator, Union, TextIO, List, Optional
import sys
from contextlib import nullcontext
from tqdm.auto import tqdm
import json
from pathlib import Path
import polars as pl

from .types import (
    EnvDict,
    ArgsMatrix,
    Includes,
    Excludes,
)
from .utils import (
    iter_dict,
    ensure_df_columns,
    ensure_df_compat,
    prepare_df_match_expr,
)
from .executors import BaseExecutor
from .providers import BaseProvider
from .run import Run


class Sweep:
    def __init__(self):
        self._executor: BaseExecutor = None

        self._args: pl.DataFrame = None

        self._provider: BaseProvider = None

    def __len__(self):
        return len(self._args)

    @property
    def tags(self) -> List[str]:
        assert self._args is not None, "Did you set .args(...) first?"

        return list(
            filter(
                None,
                [
                    f"arg:{k}"
                    for k, v in next(
                        self._args.select(
                            [pl.col(c).n_unique() for c in self._args.columns]
                        ).iter_rows(named=True)
                    ).items()
                    if v > 1
                ],
            )
        )

    def __iter__(self) -> Iterator[Run]:
        assert self._args is not None, "Did you run .args(...)?"
        assert self._executor is not None, "Did you run .executor(...)?"

        for row in self._args.iter_rows(named=True):
            run = Run(
                program=self._executor.exec,
                args={k: v for k, v in row.items() if v is not None},
                env=self._executor.env,
            )

            if self._provider:
                run = self._provider(run, sweep=self)

            yield run

    def args(self, matrix: ArgsMatrix):
        all_args = iter_dict(matrix)

        new_df = pl.DataFrame(all_args)

        if self._args is None:
            self._args = new_df
        else:
            self._args, new_df = ensure_df_compat(self._args, new_df, update=True)

            self._args = pl.concat([self._args, new_df], how="align", rechunk=True)

        self._args = self._args.unique()

        return self

    def include(self, includes: Includes):
        assert self._args is not None, "Did you set .args(...) first?"

        if not isinstance(includes, list):
            includes = [includes]

        for match_dict, include_dict in tqdm(includes, leave=False):
            ensure_df_columns(self._args, match_dict.keys())
            self._args = ensure_df_columns(self._args, include_dict.keys(), update=True)

            self._args = self._args.with_columns(
                pl.when(prepare_df_match_expr(self._args, match_dict))
                .then(pl.struct(**{k: pl.lit(v) for k, v in include_dict.items()}))
                .otherwise(pl.struct(*include_dict.keys()))
                .struct.unnest()
            )

        self._args = self._args.unique()

        return self

    def exclude(self, excludes: Excludes):
        assert self._args is not None, "Did you set .args(...) first?"

        if isinstance(excludes, dict):
            excludes = [excludes]

        for match_dict in tqdm(excludes, leave=False):
            ensure_df_columns(self._args, match_dict.keys())

            self._args = self._args.filter(
                ~prepare_df_match_expr(self._args, match_dict)
            )

        return self

    def provider(self, provider: BaseProvider):
        if self._provider is not None:
            raise ValueError(".provider() can only be set once.")

        self._provider = provider

        self.exclude(self._provider.runs(self._args.columns))

        return self

    def executor(self, executor: BaseExecutor):
        self._executor = executor

        return self

    def env(self, env_dict: EnvDict):
        assert self._executor is not None, "Did you set .executor(...) first?"

        self._executor.env = {**self._executor.env, **env_dict}

        return self

    def write_bash(self, file: Union[str, Path, TextIO] = None, delay: int = 3):
        with (
            open(file, "w")
            if isinstance(file, (str, Path))
            else nullcontext(file or sys.stdout) as file
        ):
            print("#!/usr/bin/env -S bash -l", file=file, end="\n\n")

            for run in self:
                print(str(run), file=file)
                print(f"sleep $(( RANDOM % {delay} ))", file=file, end="\n\n")

    def write_json(
        self,
        file: Union[str, Path, TextIO] = None,
        indent: Optional[int] = None,
        jsonl: bool = False,
    ):
        with (
            open(file, "w")
            if isinstance(file, (str, Path))
            else nullcontext(file or sys.stdout) as file
        ):
            if jsonl:
                for run in self:
                    print(json.dumps(run.todict()), file=file)
            else:
                print(
                    json.dumps([run.todict() for run in self], indent=indent), file=file
                )



from typing import List, Optional
from enum import Enum
from dataclasses import dataclass, field

from .base import BaseExecutor


@dataclass
class PueueExecutor(BaseExecutor):
    executable: List[str] = field(default_factory=lambda: ["puv", "python"])
    gpus: Optional[int] = field(default=None)

    class Env(str, Enum):
        GPUS = "GPUS"

    def __post_init__(self):
        if self.gpus is not None:
            self.env[self.Env.GPUS.value] = self.gpus
